﻿using System;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.ComponentModel;
using System.Diagnostics;
using System.Threading;
using AsyncReduxUtils;
using System.Runtime.Remoting.Messaging;

namespace EventBasedAsync
{
    // Define delegates for each event we want to publish
    public delegate void SlowProcessStartedEventHandler(EventArgs e);
    public delegate void ProgressChangedEventHandler(ProgressChangedEventArgs e);
    public delegate void SlowProcessCompletedEventHandler(SlowProcessCompletedEventArgs e);

    // Define SlowProcessCompletedEventArgs
    public class SlowProcessCompletedEventArgs : AsyncCompletedEventArgs
    {
        public SlowProcessCompletedEventArgs(Exception e, bool canceled, object state)
            : base(e, canceled, state)
        { }

    }

    public class SlowProcessViewModel : INotifyPropertyChanged
    {
        // Delegate on which to invoke async process
        private delegate void SlowProcessWorkerDelegate(AsyncOperation asyncOp);

        // Private dictionary to manage async op state lifetime
        private HybridDictionary _userStateToLifetime = new HybridDictionary();

        // Delegate instances for posting back to the calling thread from the invoked thread
        private SendOrPostCallback _startedCallbackDelegate;
        private SendOrPostCallback _progressCallbackDelegate;
        private SendOrPostCallback _completedCallbackDelegate;

        // Public events
        public event SlowProcessStartedEventHandler SlowProcessStarted;
        public event ProgressChangedEventHandler SlowProcessProgressChanged;
        public event SlowProcessCompletedEventHandler SlowProcessCompleted;

        public SlowProcessViewModel()
        {
            InitializeDelegates();
            SlowProcessCompleted += new SlowProcessCompletedEventHandler(SlowProcessCompletedEventHandlerMethod);
        }

        protected virtual void InitializeDelegates()
        {
            _startedCallbackDelegate = new SendOrPostCallback(SlowProcessStartedCallback);
            _progressCallbackDelegate = new SendOrPostCallback(SlowProcessProgressCallback);
            _completedCallbackDelegate = new SendOrPostCallback(SlowProcessCompletedCallback);
        }

        // Implementation

        public virtual void SlowProcessAsync(object taskId)
        {
            var token = StopwatchHelper.StartNew(TimerType.ASYNC_PROCESS_TIMER);

            // Create an AsyncOperation for taskId
            AsyncOperation asyncOp = AsyncOperationManager.CreateOperation(taskId);

            // Multiple threads will access the task dictionary, 
            //     so it must be locked to serialize access. 
            lock (_userStateToLifetime.SyncRoot)
            {
                if (_userStateToLifetime.Contains(taskId))
                {
                    throw new ArgumentException("Task ID parameter must be unique", "taskId");
                }

                Tag = taskId;
                _userStateToLifetime[taskId] = asyncOp;
            }

            // Start the asynchronous process using the delegate BeginInvoke method,
            //     passing the asycOp object
            SlowProcessWorkerDelegate workerDelegate = new SlowProcessWorkerDelegate(SlowProcess);
            workerDelegate.BeginInvoke(asyncOp, EndSlowProcessCallbackMethod, null);

            StopwatchHelper.Stop(token);
        }

        private void SlowProcess(AsyncOperation asyncOp)
        {
            var token = StopwatchHelper.StartNew(TimerType.WORKER_TIMER);

            asyncOp.Post(_startedCallbackDelegate, new EventArgs());

            bool taskCanceled = false;
            for (var i = 0; i <= 5; i++)
            {
                if (TaskCanceled(asyncOp.UserSuppliedState))
                {
                    SlowProcessCompletionMethod(null, true, asyncOp);
                    taskCanceled = true;
                    break;
                }

                Thread.Sleep(TimeSpan.FromSeconds(1));
                var progressState = new ProgressChangedEventArgs(
                    i * 20, // put into percentage terms to satisfy API
                    asyncOp.UserSuppliedState);

                asyncOp.Post(_progressCallbackDelegate, progressState);
            }

            if (!taskCanceled)
                SlowProcessCompletionMethod(null, false, asyncOp);

            StopwatchHelper.Stop(token);
        }

        private void EndSlowProcessCallbackMethod(IAsyncResult ar)
        {
            var token = StopwatchHelper.StartNew(TimerType.ASYNC_ENDINVOKE_TIMER);

            var asyncRes = (AsyncResult)ar;
            SlowProcessWorkerDelegate spd = (SlowProcessWorkerDelegate)asyncRes.AsyncDelegate;
            spd.EndInvoke(asyncRes);

            StopwatchHelper.Stop(token);
        }

        public object Tag { get; set; }

        // Started Methods

        private void SlowProcessStartedCallback(object state)
        {
            var e = new EventArgs();

            RaiseSlowProcessStarted(e);
        }

        protected void RaiseSlowProcessStarted(EventArgs e)
        {
            if (SlowProcessStarted != null)
            {
                SlowProcessStarted(e);
            }
        }

        // Progress Methods

        private void SlowProcessProgressCallback(object state)
        {
            ProgressChangedEventArgs e = state as ProgressChangedEventArgs;

            RaiseProgressChanged(e);
        }

        protected void RaiseProgressChanged(ProgressChangedEventArgs e)
        {
            if (SlowProcessProgressChanged != null)
            {
                SlowProcessProgressChanged(e);
            }
        }

        // Completion and Cancellation Methods

        // Completion methods

        // This is the method that the underlying, free-threaded  
        //     asynchronous behavior will invoke.  This will happen on 
        //     an arbitrary thread. 
        private void SlowProcessCompletionMethod(Exception exception, bool canceled, AsyncOperation asyncOp)
        {
            // If the task was not previously canceled, 
            //     remove the task from the lifetime collection. 
            if (!canceled)
            {
                lock (_userStateToLifetime.SyncRoot)
                {
                    _userStateToLifetime.Remove(asyncOp.UserSuppliedState);
                }
            }

            // Package the results of the operation in a SlowProcessCompletedEventArgs.
            var e = new SlowProcessCompletedEventArgs(exception, canceled, asyncOp.UserSuppliedState);

            // End the task. The asyncOp object is responsible  
            //     for marshalling the call to the SynchroniationContext thread.
            asyncOp.PostOperationCompleted(_completedCallbackDelegate, e);

            // Note that after the call to PostOperationCompleted,  
            //     asyncOp is no longer usable, and any attempt to use it 
            //     will cause an exception to be thrown.
        }

        // This method is invoked via the AsyncOperation object, 
        //     so it is guaranteed to be executed on the correct thread. 
        private void SlowProcessCompletedCallback(object operationState)
        {
            var e = operationState as SlowProcessCompletedEventArgs;
            RaiseSlowProcessCompleted(e);
        }

        protected void RaiseSlowProcessCompleted(SlowProcessCompletedEventArgs e)
        {
            if (SlowProcessCompleted != null)
            {
                SlowProcessCompleted(e);
            }
        }

        // Cancel methods

        // This method cancels an asynchronous process by removing the task from lifetime dictionary 
        public void CancelAsync(object taskId)
        {
            AsyncOperation asyncOp = _userStateToLifetime[taskId] as AsyncOperation;
            if (asyncOp != null)
            {
                lock (_userStateToLifetime.SyncRoot)
                {
                    _userStateToLifetime.Remove(taskId);
                }
            }
        }

        // Utility method for determining if a task has been canceled. 
        private bool TaskCanceled(object taskId)
        {
            return (_userStateToLifetime[taskId] == null);
        }

        // Completion and Cancellation Properties

        public void SlowProcessCompletedEventHandlerMethod(SlowProcessCompletedEventArgs e)
        {
            var token = StopwatchHelper.StartNew(TimerType.UPDATE_RESULT_TIMER);

            if (e.Cancelled)
                ResultData = "Slow process canceled.";
            else
                ResultData = "Slow process completed.";

            StopwatchHelper.Stop(token);
        }

        private string _resultData = "";
        public string ResultData
        {
            get
            {
                return _resultData;
            }
            set
            {
                _resultData = value;
                RaisePropertyChanged("ResultData");
            }
        }

        #region Implement INotifyPropertyChanged

        public void RaisePropertyChanged(string name)
        {
            if (PropertyChanged != null)
            {
                PropertyChanged(this, new PropertyChangedEventArgs(name));
            }
        }

        public event PropertyChangedEventHandler PropertyChanged;

        #endregion

    }

}